//==============================================================================
//  Funktionsbibliothek zur Klasse Btree: Version mit Einfgen und Lschen
//  zuletzt gendert:   07.01.01 / Kp
//==============================================================================
//  openDatabase   : Erzeugt eine neue bzw. oeffnet eine vorhandene Datenbank
//  closeDatabase  : Schliet die die Datenbank-Datei
//  insertKey      : trgt ein neues Item in die Datenbank ein
//  deleteKey      : trgt ein neues Item in die Datenbank ein
//  printDatabase  : Gibt alle Datenbank-Pages aus, die Wurzel zuerst
//  inOrderBT      : inOrder-Ausgabe der Datenbank
//
//  writePage0     : Gibt den Datenbank-Root-Record auf die Datei aus
//  allocateRec    : Ermittelt die Satznummer einer freien Page in der DB
//  loadPage       : Ldt eine Page aus der Datenbank in den Puffer BF
//  savePage       : Schreibt eine Page zurck in die Datenbank
//  searchItem     : Binre Suche nach einem Item in der aktuellen Page
//  copyItem       : Kopiert einem Item 
//  splitPage      : Spaltet eine Page auf
//  update         : fgt ein Item in einen (Teil-)Baum mit Wurzel node ein
//  releasePage    : eine Page in der Datenbank freigeben
//  compensate     : Ausgleich nach einer Lschoperation
//  delKey         : Unterprogramm fr rekursives Lschen
//  keyTransfer    : Lschen durch Schlsseltausch mit dem symm. Vorgnger
//==============================================================================

#include <string.h>
#include <iostream.h>
#include "Btree.h"

//------------------------------------------------------------------------------
//  openDatabase : 	Legt eine leere Datenbank an mit einer Page 0  oder
//                   ffnet eine vorhandene Datenbank.
//  Parameter 		   char* DBname  Name der Datenbank-Datei
//------------------------------------------------------------------------------
void Btree::openDatabase (char * DBname)
{
    // Versuchen, eine vorhandene Datenbank zu oeffnen
    if ( (DBfile=fopen(DBname,"rb+")) == NULL )
    { // Datenbank neu anlegen, wenn es noch keine gibt
      if ( (DBfile=fopen(DBname,"wb+")) == NULL )
      {  cout << "Datei  " << DBname << "  kann nicht geffnet werden !\n";
         return;
      }
      // Datenbank-Ankerfelder initialisieren und Page 0 ausgeben
      freeRec 	=  0;            	           // Anker der Freispeicherliste
      allocRec 	=  0;                     // Gre der Datenbank
      rootRec	=  0;                        // Zeiger zur Wurzel
      writePage0  ();                       // Page 0 in die Datenbank schreiben
    }
    else
    { // Anker der vorhandenen Datenbank einlesen
      fseek (DBfile, 0, SEEK_SET);
      fread (&(Page0BF.page),sizeof (DPage), 1, DBfile);
      freeRec  = Page0BF.page.sPage.FreeRec;
      allocRec = Page0BF.page.sPage.AllocRec;
      rootRec  = Page0BF.page.sPage.RootRec;
      
      // Root-Page der vorhandenen Datenbank einlesen
      fseek (DBfile, rootRec * sizeof (DPage), SEEK_SET);
      fread (&(rootBF.page), sizeof (DPage), 1, DBfile);
      rootBF.pageNum = rootRec;
    };
}

//------------------------------------------------------------------------------
//  CloseDatabase : 	Schliet die die Datenbank-Datei
//------------------------------------------------------------------------------
void Btree::closeDatabase (void)
{  fclose (DBfile); }

//------------------------------------------------------------------------------
//  writePage0 : 	Gibt den Datenbank-Root-Record auf die Datei aus
//------------------------------------------------------------------------------
void Btree::writePage0 (void)
{   Page0BF.page.sPage.AllocRec = allocRec;    // Puffer fr fwrite initialisieren
    Page0BF.page.sPage.FreeRec  = freeRec;
    Page0BF.page.sPage.RootRec  = rootRec;
    fseek (DBfile,(size_t)0,SEEK_SET);  	    // Page 0 in die DB schreiben
    fwrite (&(Page0BF.page), (size_t)sizeof(DPage), (size_t)1, DBfile);
}


//------------------------------------------------------------------------------
//  allocateRec : 	Ermittelt die Satznummer einer freien Page in der DB
//	Parameter   :	int * x :     Satznummer der freien Page
//------------------------------------------------------------------------------
void  Btree::allocateRec (int& x)
{
	Buffer  BF;

	if (freeRec == 0)  						      // Falls Freispeicher-Liste leer ist:
	{	allocRec++;                			   // Datenbankgre erhhen durch neuen Record
		x = allocRec;             			      // neue Gre in den Parameter x
		writePage0();              			   // Page 0 aktualisieren
	}
	else
	{	x = freeRec;              		 	     // Nchste freien Page zurckmelden
		fseek (DBfile, (x) * sizeof(DPage), SEEK_SET);
		fread (&(BF.page.dPage), sizeof(DPage), 1, DBfile);
		freeRec = BF.page.sPage.FreeRec;      // aus der Freispeicherliste aushngen
		writePage0();                         // Page 0 aktualisieren
	}
}

//------------------------------------------------------------------------------
//  loadPage : 	Ldt eine Page aus der Datenbank in den Puffer BF
//              Die Root-Page kann direkt aus rootBF kopiert werden
//  Parameter:  int x :  Nummer der gewnschten Page
//------------------------------------------------------------------------------
void Btree::loadPage (int x)
{
	if (x == rootRec)                          // Root-Page aus rootBF holen
	   memcpy (&BF, &rootBF, sizeof (Buffer));
	else                                       // andere Pages aus der DB holen
	{  BF.pageNum = x;                         // Page-Nummer im Puffer eintragen
	   fseek (DBfile, (size_t)(x * sizeof(DPage)), SEEK_SET);
	   fread (&(BF.page), (size_t)sizeof(DPage), (size_t)1, DBfile);
	}
}

//------------------------------------------------------------------------------
//  savePage : 	Schreibt eine Page zurck in die Datenbank
//  Parameter:  Buffer& BF  :  Puffer mit der auszugebenden Page
//------------------------------------------------------------------------------
void Btree::savePage (Buffer& BF1)
{   // falls es sich um die Root-Page handelt, rootBF aktualisieren
	if (BF1.pageNum == rootRec) memcpy (&rootBF, &BF1, sizeof(Buffer));

	// Page in die Datenbank schreiben
	fseek (DBfile, BF1.pageNum * sizeof(DPage), SEEK_SET);
	fwrite (&(BF1.page), (size_t)sizeof(DPage), (size_t)1, DBfile);
}

//------------------------------------------------------------------------------
//  searchItem :  Binre Suche nach einem Item in der aktuellen Page
//  Parameter  :  int SearchKey  : Schlssel des gesuchten Item
//				  int x	     : Position des gefundenen Item bei found=true,
//				  			   oder Position nach der es einzuordnen ist
//	              int found  : true, falls der Schlssel vorhanden ist
//------------------------------------------------------------------------------
void  Btree::searchItem (int SearchKey, int& x, bool& found)
{
	int	r, l;                                // Zeiger fr die binre Suche

	l = 0;                                     // vor das erste Item zeigen
	r =(int)(BF.page.dPage.numItems + 1);      // hinter das letzte Item zeigen
	found = false;                             // Item bisher noch nicht gefunden
	while ( (l+1 < r) && (found==false) )      // Binres Suchen
	{ x = (l+r) / 2;
	  if (SearchKey == BF.page.dPage.item[x].key)      found = true;
	  else if (SearchKey < BF.page.dPage.item[x].key)  r = x;
      else                                             l = x;
	}
	if (found == false) x=l;  // gesuchtes Item ist rechts von Pos. x  einzuordnen
}

//------------------------------------------------------------------------------
//  copyItem  : kopiert ein Item
//  Parameter :	Item& ziel
//              Item  quelle
//------------------------------------------------------------------------------
void  Btree::copyItem (Item& ziel, Item quelle)
{   ziel.key = quelle.key;
    strcpy(ziel.info, quelle.info);
    ziel.ref = quelle.ref;
}

//------------------------------------------------------------------------------
//  splitPage : Spaltet eine Page auf
//  Parameter :	Buffer *BF     :  Puffer mit der Page
//              Item   * item  :  zustzliches Item
//              int    x       :  Position, hinter die das neue Item gehrt
//------------------------------------------------------------------------------
void  Btree::splitPage (Buffer& BF, Item& item, int x)
{
	Item	SplitItem;					             // Platz fr das herauszuhebende Item
	Buffer	SplitBF;                          // Zusatzpuffer fr die neue Page
	int 	z;                                   // Laufvariable

	allocateRec(SplitBF.pageNum);	             // reserviere eine freie Page
	if (x < n) // neues Item kommt in die linke Page:
	{   // Item auf Position n nach SplitItem bringen
		copyItem (SplitItem, BF.page.dPage.item[n]);
		// Items rechts von Position x nach rechts verschieben und neues Item einsetzen
		for (z=n-1; z>=x+1; z--)
			copyItem(BF.page.dPage.item[z+1], BF.page.dPage.item[z]);
		copyItem(BF.page.dPage.item[x+1], item);
	 }
	else if (x > n) // neues Item kommt in die rechte Page
	{   // herauszunehmendes Item nach SplitItem bringen
		copyItem(SplitItem, BF.page.dPage.item[n+1]);
		// Platz schaffen fr das neue Item und dieses einsetzen
		for (z=n+2; z<=x; z++)
			copyItem(BF.page.dPage.item[z-1], BF.page.dPage.item[z]);
		copyItem(BF.page.dPage.item[x], item);
	}
	else  // neues Item kommt selbst nach SplitItem
		copyItem(SplitItem, item);

	// Puffer fr die rechte Page aufbereiten
	SplitBF.page.dPage.item[0].ref = SplitItem.ref; // erster Zeiger in der rechten Page
	SplitItem.ref = SplitBF.pageNum;           // rechte Page an SplitItem anhngen
	copyItem(item, SplitItem);         		    // SplitItem bergeben
	for (z=n+1; z <= nn; z++)                  // Items in SplitBF bernehmen
		copyItem(SplitBF.page.dPage.item[z-n], BF.page.dPage.item[z]);

	// Anzahlen gltiger Items aktualisieren
	BF.page.dPage.numItems = n;
	SplitBF.page.dPage.numItems = nn - n;

	// rechten Puffer in die Datenbank schreiben
	savePage (SplitBF);
}

//--------------------------------------------------------------------------
//  update    :  fgt ein Item in einen (Teil-)Baum mit der Wurzel node ein
//  Parameter :  int  node       :  Satznummer der Wurzel-Page
//               int&  rise      :  true, falls die Baumhhe grer wird
//               int  *risenItem :  nach oben zu verlagerndes Item
//               int  key        :  Schlssel des neuen Item
//               char *data      :  Datenfeld des neuen Item
//--------------------------------------------------------------------------
void  Btree::update (int node, int& rise, Item& risenItem, int key, char *data)
{
	int  x,z;

	if (node == 0)                             // neues Item ersetzt ein Blatt:
	{   rise            = true;                // Hhe wird ger
		risenItem.key 	= key;            	    // Schlssel eintragen
		strcpy(risenItem.info, data);           // Datenfeld eintragen
		risenItem.ref 	= 0;              	    // Zeiger initialisieren
	}
	else                                       // Wurzel des Unterbaumes laden
	{   loadPage(node);
		searchItem (key, x, found);       	    // Item im Wurzel-Record suchen
		if (found == true)                      // falls vorhanden, Datenfeld ersetzen
		{   strcpy (BF.page.dPage.item[x].info, data);
			savePage (BF);
		}
		else                                    // falls nicht vorhanden, weitersuchen
		{   risen = false;                      // Schalter fr aufsteigendes Item = false
			update (BF.page.dPage.item[x].ref, risen, risenItem, key, data);
			if (risen == true)                   // falls ein Item aufsteigen muss
			{   loadPage (node);                 // Wurzel-Page laden
				if (BF.page.dPage.numItems<nn)    // falls noch Platz in der Wurzel ist:
				{   BF.page.dPage.numItems++ ;    // Item an Position x+1 eintragen
					for (z=(int)(BF.page.dPage.numItems-1) ; z>=x+1; z--)
						copyItem(BF.page.dPage.item[z+1], BF.page.dPage.item[z]);
					copyItem(BF.page.dPage.item[x+1], risenItem);
					rise = false;
				}
				else                              // falls Wurzel voll ist, zerlegen
				{   splitPage (BF, risenItem, x);
					rise = true;                   // risenItem muss aufsteigen
				}
				savePage (BF);                    // aktuelle Wurzel sichern
			}
		}
	}
}

//--------------------------------------------------------------------------
//  insertKey : 	trgt ein neues Item in die Datenbank ein
//  Parameter 		int    key  :  Schlssel
//	                char * data :  Datenfeld
//--------------------------------------------------------------------------
void  Btree::insertKey (int key, char *data)
{
	Buffer 	newRoot;
	int 	rootSplit;
	Item 	rootItem;

	rootSplit = false;       // Schalter fr die Aufteilung der Wurzel = false
	update (rootRec, rootSplit, rootItem, key, data);
	if (rootSplit == true)       			       // neue Root wird RootItem:
	{
		allocateRec (newRoot.pageNum);          // neue Page zuweisen
		newRoot.page.dPage.numItems = 1;        // sie wird 1 Item enthalten
                                              // 1. Zeiger auf alte Root setzen
		newRoot.page.dPage.item[0].ref = rootRec;
        copyItem (newRoot.page.dPage.item[1], rootItem);
		savePage (newRoot);                     // neue Wurzel sichern
		memcpy(&rootBF, &newRoot, sizeof(BF));  // Wurzel in rootBF resident halten
		rootRec = newRoot.pageNum;              // DB-Ankerfeld RootRec aktualisieren
		writePage0();                           // Page 0 in die Datenbank ausgeben
	}
}

//-----------------------------------------------------------------------
//	releasePage  Baumknoten lschen
//  Parameter    int  z     Satznummer des zu lschenden Knotens
//-----------------------------------------------------------------------
void Btree::releasePage (int z)
{
  BF.page.sPage.FreeRec = freeRec;             // vor die Freispeicher-Liste hngen
  fseek (DBfile, z * sizeof (DPage), SEEK_SET);// Record holen
  fwrite (&(BF.page), sizeof (DPage), 1, DBfile);
  freeRec = z;                                 // neuer Anker der Freispeicher-Liste
  writePage0();
}

//--------------------------------------------------------------------------
//  compensate :  Ausgleich nach einer Lschoperation
//  Parameter  :  precedent   Vaterknoten
//                node        Sohnknoten, in dem der Unterlauf vorkam
//                path        Index des Item, das den Zeiger zum Sohn enthlt
//                underflow   true, falls es einen Unterlauf im Vaterknoten gibt
//--------------------------------------------------------------------------
void Btree::compensate (int precedent, int node, int path, bool& underflow)
{   int    neighbour;
    int    numBF2, numBF3;
    int    x, z;
    Buffer BF1, BF2, BF3;

    loadPage (node);                           // Sohnknoten in BF1 bringen
    memcpy (&BF1, &BF, sizeof(Buffer));
    loadPage (precedent);                      // Vaterknoten in BF3 bringen
    memcpy (&BF3, &BF, sizeof (Buffer));
    numBF3 = BF3.page.dPage.numItems;

    // Sohn ist >>nicht<< der rechteste Knoten
    if (path < numBF3)
    { path++;                                  // Zeiger zum rechten Nachbarn des Sohnes
      neighbour = BF3.page.dPage.item[path].ref;
      loadPage (neighbour);                    // rechten Nachbarn in BF2 bringen
      memcpy (&BF2, &BF, sizeof (Buffer));
      numBF2 = BF2.page.dPage.numItems;

      // einen Key aus dem Vaterknoten in den untergelaufenen Knoten schieben
      BF1.page.dPage.item[n-1].key = BF3.page.dPage.item[path].key;
      BF1.page.dPage.item[n-1].ref = BF2.page.dPage.item[0].ref;
      strcpy (BF1.page.dPage.item[n-1].info, BF3.page.dPage.item[path].info);

      // feststellen, ob BF2 Schluessel abgeben kann:
      x = ((numBF2-n+2) / 2);
      if (x > 0)  // BF2 kann abgeben ---> Ausgleichen
      { // die Schlssel von BF2 nach BF1 bertragen
	    for (z=1; z<x; z++)
            BF1.page.dPage.item[z+n] = BF2.page.dPage.item[z];

        // einen Key auch in den Vaterknoten bringen
	    copyItem(BF3.page.dPage.item[path], BF2.page.dPage.item[x]);
	    BF3.page.dPage.item[path].ref = neighbour;

        // im Knoten BF2 aufrumen
	    BF2.page.dPage.item[0].ref = BF2.page.dPage.item[x].ref;
	    numBF2 = (numBF2 -x);
        for (z = 1; z <= numBF2; z++)
          copyItem(BF2.page.dPage.item[z], BF2.page.dPage.item[z+x]);
        BF2.page.dPage.numItems = numBF2;
        BF1.page.dPage.numItems = (n+x-2);

        // alle beteiligten Knoten zurckschreiben
        savePage (BF1); savePage (BF2); savePage (BF3);
        underflow = false;
      }
      else  // BF2 kann nichts abgeben --> Mischen
      {  for (z=1; z<n; z++)                   // BF2 nach BF1 bertragen
            copyItem(BF1.page.dPage.item[n+z-1], BF2.page.dPage.item[z]);

         for (z=path; z<numBF3; z++)           // Schlssel in BF3 zusammenschieben
            copyItem(BF3.page.dPage.item[z], BF3.page.dPage.item[z+1]);

        // Schlselanzahlen in BF2 und BF3 korrigieren,
        // Flag fr Unterlauf im Vaterknoten setzen
        BF1.page.dPage.numItems = 2*(n-1);
		BF3.page.dPage.numItems = (numBF3-1);
		if (numBF3<n) underflow = true;
		else          underflow = false;

        // alle beteiligten Knoten zurckschreiben, BF2 lschen
        savePage (BF1); savePage (BF3);
		releasePage (neighbour);
	  }
    }
    // Sohn ist der rechteste Knoten
    else
    { // linken Nachbarn in BF2 laden
      neighbour = BF3.page.dPage.item[path-1].ref;
      loadPage (neighbour);
      memcpy (&BF2, &BF, sizeof (Buffer));

      // feststellen, ob BF2 Schluessel abgeben kann:
      numBF2 = BF2.page.dPage.numItems;
      x = ((numBF2-n+2) / 2);
      if (x > 0) // BF2 kann abgeben ---> Ausgleichen
      { // am Anfang von BF1 Platz schaffen fr n-1 Items
        for (z=n-1; z>0; z--)
           copyItem(BF1.page.dPage.item[z+x], BF1.page.dPage.item[z]);

        // Position x aus dem Vaterknoten bernehmen
		copyItem(BF1.page.dPage.item[x], BF3.page.dPage.item[path]);
		BF1.page.dPage.item[x].ref = BF1.page.dPage.item[0].ref;

        // x Items aus BF2 bernehmen in BF1
		numBF2 = (numBF2-x+1);
        for (z=x-1; z>0; z--)
           copyItem(BF1.page.dPage.item[z], BF2.page.dPage.item[z+numBF2]);
        // linkesten Zeiger in BF2 korrigieren
        BF1.page.dPage.item[0].ref = BF2.page.dPage.item[numBF2].ref;

        // einen Key aus BF2 in den Vaterknoten holen, Key-Anzahl Korrigieren
		copyItem(BF3.page.dPage.item[path], BF2.page.dPage.item[numBF2]);
		BF3.page.dPage.item[path].ref = node;
		BF2.page.dPage.numItems = numBF2-1;
		BF1.page.dPage.numItems = (n-2+x);

        // alle beteiligten Knoten zurckschreiben
		savePage (BF1); savePage (BF2); savePage (BF3);
		underflow = false;
      }
	  else  // BF2 kann nichts abgeben --> Mischen
	  { //  einen Key aus dem Vaterknoten nach BF2 bertragen
        copyItem(BF2.page.dPage.item[numBF2+1], BF3.page.dPage.item[path]);
        BF2.page.dPage.item[++numBF2].ref = BF1.page.dPage.item[0].ref;

        // Zum Mischen BF1 nach BF2 bertragen
        for (z=1; z<n-1; z++)
           BF2.page.dPage.item[z+numBF2] = BF1.page.dPage.item[z];
        // Anzahl der Knoten korrigieren
        BF2.page.dPage.numItems = nn-1;
        BF3.page.dPage.numItems = numBF3 - 1;
        if (numBF3<n) underflow = true;
        else          underflow = false;

        // alle beteiligten Knoten zurckschreiben , Knoten BF1 lschen
        savePage (BF2); savePage (BF3);
        releasePage (node);
      }
    }
}

//--------------------------------------------------------------------------
//  keyTransfer:   lscht per Schlsseltransfer mit dem symmetrischen Vorgnger
//  Parameter      int    delKey  :  zu lschender Schlssel
//--------------------------------------------------------------------------
void Btree::keyTransfer (Buffer& BF3, int node1, bool& underflow)
{   int node2;
    int numBF;
    Buffer BF1;

    loadPage (node1);                          // Wurzel des Teilbaumes laden
    memcpy (&BF1, &BF, sizeof (Buffer));       // und in lokalen Puffer bernehmen
    numBF = BF1.page.dPage.numItems;
    node2 = BF1.page.dPage.item[numBF].ref;    // rechtester Vorwrtszeiger

    if (node2 != 0) // Falls der symmetrische Vorgnger weiter rechts unten liegt
    { keyTransfer (BF3, node2, underflow);     // symm. Vorgnger weitersuchen
      if (underflow) compensate (node1, node2, numBF, underflow); // ggf. ausgleichen
    }
    else
    { // symmetrischer Vorgnger gefunden: Keytransfer durchfhren
      BF3.page.dPage.item[x].key = BF1.page.dPage.item[numBF].key;
      strcpy (BF3.page.dPage.item[x].info, BF1.page.dPage.item[numBF].info);
      // Falls der Zielknoten des Keytransfer  die Wurzel ist, rootBF updaten
      if (BF3.pageNum == rootRec) memcpy (&rootBF, &BF3, sizeof(Buffer));
      BF1.page.dPage.numItems = --numBF;       // in der Quelle des Keytransfers wird
      savePage (BF1);                          // der grte Schlssel entfernt
      if (numBF<n-1) underflow = true;         // ggf. Unterlauf nach oben melden
      else           underflow = false;
    }
}

//--------------------------------------------------------------------------
//  delKey :     lscht einen Schlssel aus der Datenbank
//  Parameter    int    delKey  :  zu lschender Schlssel
//--------------------------------------------------------------------------
void Btree::delKey (int node, bool& underflow, int key)
{   int x, y, z;
    Buffer BF3;

    if (node == 0)  underflow = false;         // leerer Baum oder Blatt
    else
    {
      loadPage (node);
      searchItem (key, x, found);
      if (found)                               // zu lschender Schlssel gefunden
      { y = BF.page.dPage.item[x-1].ref;       // zeige auf vorausgehenden Unterbaum
        if (y == 0)                            // jetzt ist dieser leer
        { BF.page.dPage.numItems=BF.page.dPage.numItems-1;  // lsche Schlssel direkt in node
          if (BF.page.dPage.numItems < n-1) underflow = true;
          else                              underflow = false;
          for (z = x; z <= BF.page.dPage.numItems; z++)
            copyItem(BF.page.dPage.item[z], BF.page.dPage.item[z+1]);
          savePage (BF);                       // schreibe Knoten wieder in die DB
        }
        else                                   // jetzt ist der U-Baum nicht leer
        { memcpy (&BF3, &BF, sizeof(Buffer));
          keyTransfer (BF3, y, underflow);     // Lschen per Keytransfer mit dem
          savePage (BF3);                      // symmetrischen Vorgnger
          if (underflow) compensate (node, y, x-1, underflow);
        }
      }
	  else                                      // Schlssel noch nicht gefunden
      { y = BF.page.dPage.item[x].ref;         // suche delKey im Unterbaum
        delKey (y, underflow, key);
        if (underflow) compensate (node, y, x, underflow);
      }                                        // gleiche ggf. aus
    }
}

//--------------------------------------------------------------------------
//  deleteKey :  lscht einen Schlssel aus der Datenbank
//  Parameter    int    delKey  :  zu lschender Schlssel
//--------------------------------------------------------------------------
 void Btree::deleteKey  (int key)
{   int  toDel;
    bool underflow;

    delKey (rootRec, underflow, key);          // lsche den Schlssel
    if ((underflow) && (rootBF.page.dPage.numItems)==0)
    {  toDel   = rootRec;                      // leere Wurzel zum Lschen vormerken
       loadPage(rootBF.page.dPage.item[0].ref);// neue Wurzel laden (in Puffer BF)
       memcpy (&rootBF, &BF, sizeof (Buffer)); // residente Root belegen
       releasePage(toDel);                     // alte Wurzel lschen
       rootRec = rootBF.pageNum;               // Satznummer der neuen Wurzel
       writePage0();                           // Page0 aktualisieren
    }
}

//--------------------------------------------------------------------------
//  inOrderDatabase :  Gibt alle Datenbank-Pages in inOrder-Folge aus
//--------------------------------------------------------------------------
void  Btree::inOrderDatabase (void)
{   cout << "\n\ninOrder-Ausgabe der Datenbank:\n";
    inOrderDB (rootRec, 1);
    cout << endl ;
}

void  Btree::inOrderDB (int x, int level)
{   int    y;
    Buffer localBF;

    if (!x==0)
    { loadPage (x);
      memcpy (&localBF, &BF, sizeof (Buffer)); // Knoten des aktuellen Level merken

      if (localBF.page.dPage.item[0].ref!= NULL)
         inOrderDB (localBF.page.dPage.item[0].ref, level+1);

      for (y=1; y<=localBF.page.dPage.numItems; y++)
      { cout << localBF.page.dPage.item[y].key << " : ";
        inOrderDB (localBF.page.dPage.item[y].ref, level+1);
      }

    }
}

//--------------------------------------------------------------------------
//  printDatabase :  Gibt alle Datenbank-Pages, die Wurzel zuerst
//--------------------------------------------------------------------------
void  Btree::printDatabase (void)
{ printDB (rootRec, 1);  cout << endl; }

void  Btree::printDB (int x, int level)
{   int    y;
    Buffer localBF;

    if (!x==0)
    { loadPage (x);
      memcpy (&localBF, &BF, sizeof (Buffer)); // Knoten des aktuellen Level merken
      if (level==1)  cout << "\nKontrollausgabe der Datenbank";
      cout << "\n\n";
      cout << "Level=" <<  level  << " : ";
      cout << "Page=" << localBF.pageNum << " : ";
      cout << "#Items="  << localBF.page.dPage.numItems;
      cout << " |   * : " << localBF.page.dPage.item[0].ref << " |   ";

      // Keys und Zeiger des aktuellen Knotens ausgeben
      for (y=1; y<=localBF.page.dPage.numItems; y++)
      cout << localBF.page.dPage.item[y].key << " : "
           << localBF.page.dPage.item[y].ref << " |  ";

      // Unterbume ausgeben
      for (y=0; y<=localBF.page.dPage.numItems; y++)
          printDB (localBF.page.dPage.item[y].ref, level+1);
    }
}






